;; Generated package description from flx.el  -*- no-byte-compile: t -*-
(define-package "flx" "0.6.2" "fuzzy matching with good sorting" '((cl-lib "0.3")) :commit "29e3664b75618847eaaf821713a74eb3c897a124" :url "https://github.com/lewang/flx")
